@@ -1,8 +1,6 @@
 /*
  * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
  *
- * SPDX-License-Identifier: MPL-2.0
- *
  * This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, you can obtain one at https://mozilla.org/MPL/2.0/.
@@ -68,7 +66,7 @@ static dns_dbimplementation_t rbtimp;
 
 static void
 initialize(void) {
-	isc_rwlock_init(&implock, 0, 0);
+	RUNTIME_CHECK(isc_rwlock_init(&implock, 0, 0) == ISC_R_SUCCESS);
 
 	rbtimp.name = "rbt";
 	rbtimp.create = dns_rbtdb_create;
@@ -80,7 +78,7 @@ initialize(void) {
 	ISC_LIST_APPEND(implementations, &rbtimp, link);
 }
 
-static dns_dbimplementation_t *
+static inline dns_dbimplementation_t *
 impfind(const char *name) {
 	dns_dbimplementation_t *imp;
 
@@ -684,8 +682,7 @@ dns_db_findrdataset(dns_db_t *db, dns_dbnode_t *node, dns_dbversion_t *version,
 
 isc_result_t
 dns_db_allrdatasets(dns_db_t *db, dns_dbnode_t *node, dns_dbversion_t *version,
-		    unsigned int options, isc_stdtime_t now,
-		    dns_rdatasetiter_t **iteratorp) {
+		    isc_stdtime_t now, dns_rdatasetiter_t **iteratorp) {
 	/*
 	 * Make '*iteratorp' an rdataset iteratator for all rdatasets at
 	 * 'node' in version 'version' of 'db'.
@@ -694,8 +691,7 @@ dns_db_allrdatasets(dns_db_t *db, dns_dbnode_t *node, dns_dbversion_t *version,
 	REQUIRE(DNS_DB_VALID(db));
 	REQUIRE(iteratorp != NULL && *iteratorp == NULL);
 
-	return ((db->methods->allrdatasets)(db, node, version, options, now,
-					    iteratorp));
+	return ((db->methods->allrdatasets)(db, node, version, now, iteratorp));
 }
 
 isc_result_t
@@ -847,10 +843,10 @@ dns_db_adjusthashsize(dns_db_t *db, size_t size) {
 }
 
 void
-dns_db_settask(dns_db_t *db, isc_task_t *task, isc_task_t *prunetask) {
+dns_db_settask(dns_db_t *db, isc_task_t *task) {
 	REQUIRE(DNS_DB_VALID(db));
 
-	(db->methods->settask)(db, task, prunetask);
+	(db->methods->settask)(db, task);
 }
 
 isc_result_t
@@ -1015,7 +1011,7 @@ dns_db_rpz_ready(dns_db_t *db) {
 	return ((db->methods->rpz_ready)(db));
 }
 
-/*
+/**
  * Attach a notify-on-update function the database
  */
 isc_result_t
@@ -1026,16 +1022,6 @@ dns_db_updatenotify_register(dns_db_t *db, dns_dbupdate_callback_t fn,
 	REQUIRE(db != NULL);
 	REQUIRE(fn != NULL);
 
-	for (listener = ISC_LIST_HEAD(db->update_listeners); listener != NULL;
-	     listener = ISC_LIST_NEXT(listener, link))
-	{
-		if ((listener->onupdate == fn) &&
-		    (listener->onupdate_arg == fn_arg))
-		{
-			return (ISC_R_SUCCESS);
-		}
-	}
-
 	listener = isc_mem_get(db->mctx, sizeof(dns_dbonupdatelistener_t));
 
 	listener->onupdate = fn;
@@ -1058,8 +1044,7 @@ dns_db_updatenotify_unregister(dns_db_t *db, dns_dbupdate_callback_t fn,
 	     listener = ISC_LIST_NEXT(listener, link))
 	{
 		if ((listener->onupdate == fn) &&
-		    (listener->onupdate_arg == fn_arg))
-		{
+		    (listener->onupdate_arg == fn_arg)) {
 			ISC_LIST_UNLINK(db->update_listeners, listener, link);
 			isc_mem_put(db->mctx, listener,
 				    sizeof(dns_dbonupdatelistener_t));
@@ -1104,28 +1089,6 @@ dns_db_getservestalettl(dns_db_t *db, dns_ttl_t *ttl) {
 	return (ISC_R_NOTIMPLEMENTED);
 }
 
-isc_result_t
-dns_db_setservestalerefresh(dns_db_t *db, uint32_t interval) {
-	REQUIRE(DNS_DB_VALID(db));
-	REQUIRE((db->attributes & DNS_DBATTR_CACHE) != 0);
-
-	if (db->methods->setservestalerefresh != NULL) {
-		return ((db->methods->setservestalerefresh)(db, interval));
-	}
-	return (ISC_R_NOTIMPLEMENTED);
-}
-
-isc_result_t
-dns_db_getservestalerefresh(dns_db_t *db, uint32_t *interval) {
-	REQUIRE(DNS_DB_VALID(db));
-	REQUIRE((db->attributes & DNS_DBATTR_CACHE) != 0);
-
-	if (db->methods->getservestalerefresh != NULL) {
-		return ((db->methods->getservestalerefresh)(db, interval));
-	}
-	return (ISC_R_NOTIMPLEMENTED);
-}
-
 isc_result_t
 dns_db_setgluecachestats(dns_db_t *db, isc_stats_t *stats) {
 	REQUIRE(dns_db_iszone(db));
