@@ -1,8 +1,6 @@
 /*
  * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
  *
- * SPDX-License-Identifier: MPL-2.0
- *
  * This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, you can obtain one at https://mozilla.org/MPL/2.0/.
@@ -93,7 +91,7 @@ fromtext_in_atma(ARGS_FROMTEXT) {
 				lastwasperiod = true;
 				continue;
 			}
-			if (!isdigit((unsigned char)sr->base[0])) {
+			if ((sr->base[0] < '0') || (sr->base[0] > '9')) {
 				RETTOK(DNS_R_SYNTAX);
 			}
 			RETERR(mem_tobuffer(target, sr->base, 1));
@@ -159,7 +157,7 @@ fromwire_in_atma(ARGS_FROMWIRE) {
 	if (region.base[0] == 1) {
 		unsigned int i;
 		for (i = 1; i < region.length; i++) {
-			if (!isdigit((unsigned char)region.base[i])) {
+			if (region.base[i] < '0' || region.base[i] > '9') {
 				return (DNS_R_FORMERR);
 			}
 		}
