@@ -3,7 +3,7 @@
  *
  * This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
- * file, You can obtain one at http://mozilla.org/MPL/2.0/.
+ * file, you can obtain one at https://mozilla.org/MPL/2.0/.
  *
  * See the COPYRIGHT file distributed with this work for additional
  * information regarding copyright ownership.
@@ -1502,9 +1502,9 @@ cleanup_task:
 	dns_rbt_destroy(&zones->rbt);
 
 cleanup_rbt:
-	isc_refcount_decrement(&zones->irefs);
+	isc_refcount_decrementz(&zones->irefs);
 	isc_refcount_destroy(&zones->irefs);
-	isc_refcount_decrement(&zones->refs);
+	isc_refcount_decrementz(&zones->refs);
 	isc_refcount_destroy(&zones->refs);
 
 	isc_mutex_destroy(&zones->maint_lock);
@@ -1587,7 +1587,7 @@ cleanup_ht:
 	isc_timer_detach(&zone->updatetimer);
 
 cleanup_timer:
-	isc_refcount_decrement(&zone->refs);
+	isc_refcount_decrementz(&zone->refs);
 	isc_refcount_destroy(&zone->refs);
 
 	isc_mem_put(rpzs->mctx, zone, sizeof(*zone));
@@ -1787,7 +1787,7 @@ finish_update(dns_rpz_zone_t *rpz) {
 	/*
 	 * If there's an update pending, schedule it.
 	 */
-	if (rpz->updatepending == true) {
+	if (rpz->updatepending) {
 		if (rpz->min_update_interval > 0) {
 			uint64_t defer = rpz->min_update_interval;
 			char dname[DNS_NAME_FORMATSIZE];
@@ -1858,6 +1858,14 @@ cleanup_quantum(isc_task_t *task, isc_event_t *event) {
 
 	name = dns_fixedname_initname(&fname);
 
+	LOCK(&rpz->rpzs->maint_lock);
+
+	/* Check that we aren't shutting down. */
+	if (rpz->rpzs->zones[rpz->num] == NULL) {
+		UNLOCK(&rpz->rpzs->maint_lock);
+		goto cleanup;
+	}
+
 	for (result = isc_ht_iter_first(iter);
 	     result == ISC_R_SUCCESS && count++ < DNS_RPZ_QUANTUM;
 	     result = isc_ht_iter_delcurrent_next(iter))
@@ -1886,6 +1894,7 @@ cleanup_quantum(isc_task_t *task, isc_event_t *event) {
 			       iter, rpz, NULL, NULL);
 		nevent = &rpz->updateevent;
 		isc_task_send(rpz->rpzs->updater, &nevent);
+		UNLOCK(&rpz->rpzs->maint_lock);
 		return;
 	} else if (result == ISC_R_NOMORE) {
 		isc_ht_t *tmpht = NULL;
@@ -1898,11 +1907,14 @@ cleanup_quantum(isc_task_t *task, isc_event_t *event) {
 		rpz->nodes = rpz->newnodes;
 		rpz->newnodes = tmpht;
 
+		UNLOCK(&rpz->rpzs->maint_lock);
 		finish_update(rpz);
 		dns_name_format(&rpz->origin, domain, DNS_NAME_FORMATSIZE);
 		isc_log_write(dns_lctx, DNS_LOGCATEGORY_GENERAL,
 			      DNS_LOGMODULE_MASTER, ISC_LOG_INFO,
 			      "rpz: %s: reload done", domain);
+	} else {
+		UNLOCK(&rpz->rpzs->maint_lock);
 	}
 
 	/*
@@ -1946,6 +1958,14 @@ update_quantum(isc_task_t *task, isc_event_t *event) {
 
 	dns_name_format(&rpz->origin, domain, DNS_NAME_FORMATSIZE);
 
+	LOCK(&rpz->rpzs->maint_lock);
+
+	/* Check that we aren't shutting down. */
+	if (rpz->rpzs->zones[rpz->num] == NULL) {
+		UNLOCK(&rpz->rpzs->maint_lock);
+		goto cleanup;
+	}
+
 	while (result == ISC_R_SUCCESS && count++ < DNS_RPZ_QUANTUM) {
 		char namebuf[DNS_NAME_FORMATSIZE];
 		dns_rdatasetiter_t *rdsiter = NULL;
@@ -1988,6 +2008,7 @@ update_quantum(isc_task_t *task, isc_event_t *event) {
 			continue;
 		}
 
+		dns_name_downcase(name, name, NULL);
 		result = isc_ht_add(rpz->newnodes, name->ndata, name->length,
 				    rpz);
 		if (result != ISC_R_SUCCESS) {
@@ -2046,6 +2067,7 @@ update_quantum(isc_task_t *task, isc_event_t *event) {
 			       rpz, NULL, NULL);
 		nevent = &rpz->updateevent;
 		isc_task_send(rpz->rpzs->updater, &nevent);
+		UNLOCK(&rpz->rpzs->maint_lock);
 		return;
 	} else if (result == ISC_R_NOMORE) {
 		/*
@@ -2060,12 +2082,16 @@ update_quantum(isc_task_t *task, isc_event_t *event) {
 			       NULL, rpz, NULL, NULL);
 		nevent = &rpz->updateevent;
 		isc_task_send(rpz->rpzs->updater, &nevent);
+		UNLOCK(&rpz->rpzs->maint_lock);
 		return;
 	}
 
 	/*
 	 * If we're here, something went wrong, so clean up.
 	 */
+	UNLOCK(&rpz->rpzs->maint_lock);
+
+cleanup:
 	if (rpz->updbit != NULL) {
 		dns_dbiterator_destroy(&rpz->updbit);
 	}
@@ -2199,10 +2225,11 @@ rpz_detach(dns_rpz_zone_t **rpzp) {
 		if (dns_name_dynamic(&rpz->cname)) {
 			dns_name_free(&rpz->cname, rpzs->mctx);
 		}
-		if (rpz->dbversion != NULL) {
-			dns_db_closeversion(rpz->db, &rpz->dbversion, false);
-		}
 		if (rpz->db != NULL) {
+			if (rpz->dbversion != NULL) {
+				dns_db_closeversion(rpz->db, &rpz->dbversion,
+						    false);
+			}
 			dns_db_updatenotify_unregister(
 				rpz->db, dns_rpz_dbupdate_callback, rpz);
 			dns_db_detach(&rpz->db);
@@ -2215,9 +2242,14 @@ rpz_detach(dns_rpz_zone_t **rpzp) {
 			if (rpz->newnodes != NULL) {
 				isc_ht_destroy(&rpz->newnodes);
 			}
-			dns_db_closeversion(rpz->updb, &rpz->updbversion,
-					    false);
-			dns_db_detach(&rpz->updb);
+			if (rpz->updb != NULL) {
+				if (rpz->updbversion != NULL) {
+					dns_db_closeversion(rpz->updb,
+							    &rpz->updbversion,
+							    false);
+				}
+				dns_db_detach(&rpz->updb);
+			}
 		}
 
 		isc_timer_reset(rpz->updatetimer, isc_timertype_inactive, NULL,
@@ -2248,8 +2280,7 @@ dns_rpz_detach_rpzs(dns_rpz_zones_t **rpzsp) {
 	*rpzsp = NULL;
 
 	if (isc_refcount_decrement(&rpzs->refs) == 1) {
-		isc_task_destroy(&rpzs->updater);
-
+		LOCK(&rpzs->maint_lock);
 		/*
 		 * Forget the last of view's rpz machinery after
 		 * the last reference.
@@ -2262,6 +2293,7 @@ dns_rpz_detach_rpzs(dns_rpz_zones_t **rpzsp) {
 				rpz_detach(&rpz);
 			}
 		}
+		UNLOCK(&rpzs->maint_lock);
 		rpz_detach_rpzs(&rpzs);
 	}
 }
@@ -2285,6 +2317,7 @@ rpz_detach_rpzs(dns_rpz_zones_t **rpzsp) {
 		if (rpzs->rbt != NULL) {
 			dns_rbt_destroy(&rpzs->rbt);
 		}
+		isc_task_destroy(&rpzs->updater);
 		isc_mutex_destroy(&rpzs->maint_lock);
 		isc_rwlock_destroy(&rpzs->search_lock);
 		isc_refcount_destroy(&rpzs->refs);
@@ -2714,6 +2747,7 @@ dns_rpz_find_name(dns_rpz_zones_t *rpzs, dns_rpz_type_t rpz_type,
 	nmnode = NULL;
 	result = dns_rbt_findnode(rpzs->rbt, trig_name, NULL, &nmnode, &chain,
 				  DNS_RBTFIND_EMPTYDATA, NULL, NULL);
+
 	switch (result) {
 	case ISC_R_SUCCESS:
 		nm_data = nmnode->data;
@@ -2728,7 +2762,42 @@ dns_rpz_find_name(dns_rpz_zones_t *rpzs, dns_rpz_type_t rpz_type,
 
 	case DNS_R_PARTIALMATCH:
 		i = chain.level_matches;
-		while (i >= 0 && (nmnode = chain.levels[i]) != NULL) {
+		nmnode = chain.levels[chain.level_matches];
+
+		/*
+		 * Whenever an exact match is found by dns_rbt_findnode(),
+		 * the highest level node in the chain will not be put into
+		 * chain->levels[] array, but instead the chain->end
+		 * pointer will be adjusted to point to that node.
+		 *
+		 * Suppose we have the following entries in a rpz zone:
+		 *   example.com     CNAME rpz-passthru.
+		 *   *.example.com   CNAME rpz-passthru.
+		 *
+		 * A query for www.example.com would result in the
+		 * following chain object returned by dns_rbt_findnode():
+		 *   chain->level_count = 2
+		 *   chain->level_matches = 2
+		 *   chain->levels[0] = .
+		 *   chain->levels[1] = example.com
+		 *   chain->levels[2] = NULL
+		 *   chain->end = www
+		 *
+		 * Since exact matches only care for testing rpz set bits,
+		 * we need to test for rpz wild bits through iterating the
+		 * nodechain, and that includes testing the rpz wild bits
+		 * in the highest level node found. In the case of an exact
+		 * match, chain->levels[chain->level_matches] will be NULL,
+		 * to address that we must use chain->end as the start
+		 * point, then iterate over the remaining levels in the
+		 * chain.
+		 */
+		if (nmnode == NULL) {
+			--i;
+			nmnode = chain.end;
+		}
+
+		while (nmnode != NULL) {
 			nm_data = nmnode->data;
 			if (nm_data != NULL) {
 				if (rpz_type == DNS_RPZ_TYPE_QNAME) {
@@ -2737,7 +2806,13 @@ dns_rpz_find_name(dns_rpz_zones_t *rpzs, dns_rpz_type_t rpz_type,
 					found_zbits |= nm_data->wild.ns;
 				}
 			}
-			i--;
+
+			if (i >= 0) {
+				nmnode = chain.levels[i];
+				--i;
+			} else {
+				break;
+			}
 		}
 		break;
 
